#=======================================================================
#	$Id: Cfw.pm,v 1.4 2006/09/13 08:24:30 pythontech Exp $
#	Component framework for perl
#	Copyright (C) 2005-2006  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#-----------------------------------------------------------------------
#	Declare deployment descriptor, including objects with their
#	class and attributes.  May include references to other objects
#	by path.
#=======================================================================

#-----------------------------------------------------------------------
#	In configuration module:
#	 use PythonTech::Cfw;
#	 cfwset(foo => bar,
#		baz => cfwobj('Baz::Thing'),
#		       bazfile => 'baz.conf',
#		       stuff => 42),
#		quux => cfwobj('Quux::Object',
#			       bazzer => cfwref('baz')));
#	In executable:
#	 my $baz = cfw('baz');
#-----------------------------------------------------------------------
package PythonTech::Cfw;
use strict;
use Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(cfwset cfwobj cfwref cfw cfwexists);

my %config;			# Configuration tree
my %inst;			# Instantiated objects

#-----------------------------------------------------------------------
#	Define configuration values
#-----------------------------------------------------------------------
sub cfwset {
    while (my($key,$val) = splice(@_,0,2)) {
	my @path = split m!/!, $key;
	my $last = pop @path;
	my $pos = \%config;
	foreach my $segment (@path) {
	    $pos->{$segment} = {} unless exists $pos->{$segment};
	    $pos = $pos->{$segment};
	}
	$pos->{$last} = $val;
    }
}

#-----------------------------------------------------------------------
#	Look up config database, translating nodes (i.e instantiating
#	objects) as we go.
#	The 'x' version raises an exception if the value is not found.
#-----------------------------------------------------------------------
sub cfw {
    my($path) = @_;
    if (exists $inst{$path}) {
	return $inst{$path};
    } else {
	return &_get("",\%config,$path);
    }
}

sub cfwx {
    my($path) = @_;
    if (exists $inst{$path}) {
	return $inst{$path};
    } else {
	return &_get("",\%config,$path,1);
    }
}

sub _get {
    my($prefix,$conf,$path,$die) = @_;
    my($seg,$rest) = $path =~ m!^([^/]*)(?:/(.*))?$!;
    if (! exists $conf->{$seg}) {
	die "Cfw: bad path $prefix$seg\n" if $die;
	return undef;
    } elsif (defined $rest) {
	return _get("$prefix$seg/",$conf->{$seg},$rest,$die);
    } else {
	my $obj = &_xlate($conf->{$seg},$die);
	$inst{"$prefix$path"} = $obj;
	return $obj;
    }
}

#-----------------------------------------------------------------------
#	Check if a path exists in the configuration tree, without
#	translating (i.e. instantiating) it.  Returns 1 if it exists,
#	else undef.
#-----------------------------------------------------------------------
sub cfwexists {
    my($key) = @_;
    my @path = 
    my $pos = \%config;
    foreach my $segment (split m!/!, $key) {
	return unless ref($pos) eq 'HASH';
	return unless exists $pos->{$segment};
	$pos = $pos->{$segment};
    }
    return 1;
}

#-----------------------------------------------------------------------
#	Within definition, declare an object to be instantiated on
#	demand
#-----------------------------------------------------------------------
sub cfwobj {
    my(@clsprops) = @_;
    return bless \@clsprops, 'Cfw::Obj';
}

#-----------------------------------------------------------------------
#	Within definition, declare reference to another named config
#	item
#-----------------------------------------------------------------------
sub cfwref {
    my($path) = @_;
    return bless \$path, 'Cfw::Ref';
}

#-----------------------------------------------------------------------
#	Translate from configuration space to instantiated values
#-----------------------------------------------------------------------
sub _xlate {
    my($obj,$die) = @_;
    if (ref($obj) eq 'Cfw::Ref') {
	return $die ? &cfwx($$obj) : &cfw($$obj);
    } elsif (ref($obj) eq 'Cfw::Obj') {
	my($cls,@props) = @$obj;
	eval "use $cls";
	die $@ if $@;
	my @conf = map {&_xlate($_,$die)} @props;
	return $cls->new(@conf);
    } elsif (ref($obj) eq 'HASH') {
	my %new;
	while (my($key,$val) = each %$obj) {
	    $new{$key} = &_xlate($val,$die);
	}
	return \%new;
    } elsif (ref($obj) eq 'ARRAY') {
	my @new = map {&_xlate($_,$die)} @$obj;
	return \@new;
    } elsif (ref($obj) eq 'CODE') {
	my($new) = &$obj;
	return $new;
    } else {
	return $obj;
    }
}

# For debugging
sub all {
    return \%config;
}

1;
